/****************************************************************************
 *
 * Pocketop Wireless Keyboard(TM) add-on driver for linux-Xwindow. v0.3.3
 * Copyright GPL Licence. 
 * 2005.05.14 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

struct keymapping {
    unsigned char hwcode;        /* Input keycode from IR Keyboard to Linux
				    IrDa port( = this program) */
    unsigned int xcode_nml;      /* Output keycode from this driver to X
				    server according to xmodmap */ 
    unsigned int xcode_num[3];   /* xcode with Num key */
    unsigned int xcode_punct[3]; /* xcode with Punctuation key */
    unsigned int xcode_func[3];  /* xcode with Function key */
};

/* This is default keymapping of debian woody */ 
struct keymapping KeyTable[54] = {
    {0x69, 23, {0}, {0}, {9,0,0}},    /* Tab Tab:ISO_Left_Tab none none Escape */
    {0x01, 24, {10}, {50,10}, {0}},   /* Q q:Q 1 exclam none */ 
    {0x09, 25, {11}, {50,11}, {0}},   /* W w:W 2 at none */
    {0x11, 26, {12}, {50,12}, {0}},   /* E e:E 3 numbersign none */ 
    {0x19, 27, {13}, {50,13}, {0}},   /* R r:R 4 dollar none */  
    {0x21, 28, {14}, {50,14}, {0}},   /* T t:T 5 percent none */
    {0x29, 29, {15}, {50,15}, {0}},   /* Y y:Y 6 asciicircum none */
    {0x31, 30, {16}, {50,16}, {0}},   /* U u:U 7 ampersand none */
    {0x39, 31, {17}, {50,17}, {0}},   /* I i:I 8 asterisk none */
    {0x41, 32, {18}, {50,18}, {0}},   /* O o:O 9 parenleft none */
    {0x49, 33, {19}, {50,19}, {0}},   /* P p:P 0 parenright none */
    {0x51, 20, {61}, {50,21}, {0}},   /* Minus minus:underscore slash plus
				       none */
    {0x59, 22, {0}, {0}, {0}},        /* BackSp BackSpace:Terminate_Server
				       none none none */
    {0x77, 66, {0}, {0}, {0}},        /* Cap_Lk Caps_Lock:none none none
				       none */
    {0x07, 38, {0}, {49}, {0}},       /* A a:A none grave none */
    {0x0f, 39, {0}, {50,49}, {0}},    /* S s:S none asciitilde none */
    {0x17, 40, {0}, {61}, {0}},       /* D d:D none slash none */
    {0x1f, 41, {0}, {50,61}, {0}},    /* F f:F none slash none */
    {0x27, 42, {0}, {34}, {0}},       /* G g:G none bracketleft none */
    {0x2f, 43, {0}, {35}, {0}},       /* H h:H none bracketright none */
    {0x32, 44, {13}, {50,34}, {0}},   /* J j:J 4 braceleft none */
    {0x3a, 45, {14}, {50,35}, {0}},   /* K k:K 5 braceright none */
    {0x42, 46, {15}, {51}, {0}},      /* L l:L 6 backslash none */
    {0x4a, 47, {50,17}, {50,51}, {0}},  /* Semicolon semicolon:colon
					 asterisk bar none */
    {0x52, 48, {0}, {21}, {0}},       /* Apostrophe apostrophe:quotedbl
				       none euqal none */
    {0x5a, 36, {0}, {0}, {0}},        /* Enter return:none none none none */
    {0x65, 50, {0}, {0}, {0}},        /* Shift-L Shift_L:none none none
				       none */
    {0x03, 52, {0}, {0}, {0}},        /* Z z:Z none none none */
    {0x0b, 53, {0}, {0}, {0}},        /* X x:X none none none */
    {0x13, 54, {0}, {0}, {0}},        /* C c:C none none none */ 
    {0x1b, 55, {0}, {0}, {0}},        /* V v:V none none none */
    {0x23, 56, {0}, {0}, {0}},        /* B b:B none none none */
    {0x33, 57, {0}, {0}, {0}},        /* N n:N none none none */
    {0x3b, 58, {10}, {0}, {0}},       /* M m:M 1 none none */ 
    {0x43, 59, {11}, {0}, {0}},       /* Comma comma:less 2 none none */
    {0x4b, 60, {12}, {0}, {0}},       /* Period period:greater 3 none none */
    {0x53, 98, {20}, {0}, {99}},      /* Up Up:none minus none Prior */  
    {0x7d, 62, {0}, {0}, {0}},        /* Shift-R Shift_R:none none none
				       none */ 
    {0x6d, 37, {0}, {0}, {0}},        /* Ctrl Control_L:none  none none
				       none */ 
    {0x76, 0, {0}, {0}, {0}},         /* Punct none:none none none none */
    {0x75, 9, {0}, {0}, {0}},         /* Cmd Escape:none none none none */
    {0x6e, 0, {0}, {0}, {0}},         /* Num none:none none none none */
    {0x66, 64, {0}, {0}, {0}},        /* Alt Alt_L:Meta_L none none none */ 
    {0x2b, 65, {0}, {0}, {0}},        /* Space-L space:none none none none */
    {0x34, 65, {0}, {0}, {0}},        /* Space-R space:none none none none */
    {0x3c, 0, {19}, {0}, {0}},        /* Function none:none 0 none none */
    {0x44, 100, {60}, {0}, {97}},     /* Left Left:none period none Home */ 
    {0x4c, 104, {50,21}, {0}, {105}}, /* Down Down:none plus none Next */
    {0x54, 102, {21}, {0}, {103}},    /* Right Right:none equal none End */
    {0x5b, 107, {0}, {0}, {0}},       /* Del Delete:none none none none */ 
    {0x40, 0, {0}, {0}, {0}},         /* Date none:none none none none */ 
    {0x48, 0, {0}, {0}, {0}},         /* Address none:none none none none */
    {0x50, 0, {0}, {0}, {0}},         /* ToDo none:none none none none */
    {0x58, 0, {0}, {0}, {0}}          /* Memo none:none none none none */
};

int fd;
unsigned int idx;
unsigned char hwcode = 128;           /* initial state is break */
unsigned char prev_key;                     
unsigned int nrepeat = 0;
unsigned int num_FLAG = 0;
unsigned int punct_FLAG = 0;
unsigned int func_FLAG = 0;
unsigned int typematic_delay = 1000;   /* milliseconds */
unsigned int typematic_rate = 100;     /* milliseconds */
unsigned int auto_off_time = 300;      /* seconds */    

Display *d;

struct itimerval auto_off;
struct itimerval typematic, typematic0;

void read_rcfile() {
    int i;
    FILE *fp;
    char dmy[255];
    
    fp = fopen("/etc/pktrc", "r");
    if ( fp == NULL ) {
	fprintf(stderr, "Cannot open /etc/pktrc.\n");
	fprintf(stderr, "Internal key definition is applied.\n");
	return;
    }

    while ( (fgets(dmy, sizeof(dmy), fp)) != NULL ) {

	if ( dmy[0] == '#' ) continue;
	else if ( dmy[0] == 'd' ) {
	    sscanf(dmy, "%*s %d", &typematic_delay);
	} else if ( dmy[0] == 'r' ) {
	    sscanf(dmy, "%*s %d", &typematic_rate);
	} else if ( dmy[0] == 'a' ) {
	    sscanf(dmy, "%*s %d", &auto_off_time);
	} else if ( isdigit( dmy[0] ) != 0 ) {
	    for ( i=0; i <= 53; i++) {
		sscanf(dmy,"%x %d %d:%d:%d %d:%d:%d %d:%d:%d", 
		       &KeyTable[i].hwcode, 
		       &KeyTable[i].xcode_nml,
		       &KeyTable[i].xcode_num[0],
		       &KeyTable[i].xcode_num[1],
		       &KeyTable[i].xcode_num[2],
		       &KeyTable[i].xcode_punct[0],
		       &KeyTable[i].xcode_punct[1],
		       &KeyTable[i].xcode_punct[2],
		       &KeyTable[i].xcode_func[0],
		       &KeyTable[i].xcode_func[1],
		       &KeyTable[i].xcode_func[2]);
		fgets(dmy, sizeof(dmy), fp);
	    } 
	}
    }
    fclose(fp);
}

void send_xcodes( unsigned int *ptr, unsigned char msb) {

    int j;
    unsigned int *ptemp;

    if ( msb == 0 ) {       /* send xcodes in proper order */
	while ( *ptr != 0 ) {
	    XTestFakeKeyEvent(d, *ptr++, (1^msb), 0);
	    XFlush(d);
	} 
    } else {                /* send xcodes in reverse order */
	ptemp = ptr;

	for ( j = 0; *ptr != 0; j++) ptr++;
	for ( --j ; j >= 0; j-- ) {
	    XTestFakeKeyEvent(d, *(ptemp+j), (1^msb), 0); 
	    XFlush(d);
	}
    }
}

void translate( unsigned int i, unsigned char ch ) {
    
    unsigned char msb;   /* 0 : key pressed  */ 
			 /* 1 : key released */

    msb = (ch & 128) >> 7;   
    ch &= 127;  

    if ( ch == 0x6e ) {         /* Num key */  
	num_FLAG = (1^msb);
    } else if ( ch == 0x76 ) {  /* Punctuation Key */
	punct_FLAG = (1^msb);
    } else if ( ch == 0x3c ) {  /* function Key */
	func_FLAG = (1^msb);
    } else if ( num_FLAG == 1 ) {
	send_xcodes(&KeyTable[i].xcode_num[0], msb);		    
    } else if ( punct_FLAG == 1 ) {
	send_xcodes(&KeyTable[i].xcode_punct[0], msb);		    
    } else if ( func_FLAG == 1 ) {
	if ( ch == 0x59 ) {              /* pocketop off */
	    close(fd);
	    exit(EXIT_SUCCESS);
	};
	send_xcodes(&KeyTable[i].xcode_func[0], msb);		    
    } else {
	XTestFakeKeyEvent(d, KeyTable[i].xcode_nml, (1^msb), 0);
	XFlush(d);
    }
}

void process_hwcode( unsigned int i, unsigned char ch ) {

    /* ignore duplicated break code of pocketop */
    if ( ch == prev_key ) return;       

    translate( i, ch );
    if ( (ch & 128) == 0) {
	setitimer(ITIMER_REAL, &typematic, NULL);
    } else {
	setitimer(ITIMER_REAL, &auto_off, NULL);
    }
    prev_key = ch;	
    return;
}

void alarm_inthandler() {
    if ( (128 & hwcode) == 0 ) {     /* if key make, repeat */
	nrepeat++;
	translate( idx, hwcode );
	if ( nrepeat >= 254 ) {      /* bad coverage. forgive me */
	    nrepeat = 0;
	    process_hwcode( idx, (hwcode | 128) );  /* send break */
	    close(fd);
	    exit(EXIT_FAILURE);
	} 
    } else {
	close(fd);
	exit(EXIT_SUCCESS);
    }
}
	
int main(int argc, char *argv[]) {

    int i;
    int nread;
    char *device = "/dev/ttyS1";

    struct termios oldtio, newtio;
    struct sigaction newhandler;
    sigset_t blocked;

    /* signal install */
    newhandler.sa_handler = alarm_inthandler;
    newhandler.sa_flags = SA_RESTART;
    sigemptyset(&blocked);
    sigaddset(&blocked,SIGALRM);

    newhandler.sa_mask = blocked;    
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
        perror("sigaction install error\n");
        exit(EXIT_FAILURE);
    }
    
    /* Opening irda port device */
    if ( argc >= 2 ) device = argv[1];     
    fd = open(device, O_RDONLY | O_NOCTTY );
    if (fd == -1)  {
    	fprintf(stderr, "Error opening %s\n", device);
    	exit(EXIT_FAILURE);
    }
    
    read_rcfile();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(fd, &oldtio);        
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;        
    newtio.c_lflag = 0;      
    newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 char received */       
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
               
    /* Timer setting */
    typematic.it_interval.tv_sec = typematic_rate / 1000;
    typematic.it_interval.tv_usec = (typematic_rate % 1000) * 1000;
    typematic.it_value.tv_sec = typematic_delay / 1000;
    typematic.it_value.tv_usec = (typematic_delay % 1000) * 1000; 

    typematic0.it_value.tv_sec = 0;
    typematic0.it_value.tv_usec = 0;

    auto_off.it_value.tv_sec = auto_off_time;
    auto_off.it_value.tv_usec = 0;
    

    d = XOpenDisplay(NULL);    
    setitimer(ITIMER_REAL, &auto_off, NULL);

    while (1) {
	nread = read(fd, &hwcode, 1);	  
	if ( nread > 0 ) {
	 
	    setitimer(ITIMER_REAL, &auto_off, NULL);

	    for ( i = 0; i <= 53; i++ ) {
		if ( (hwcode & 127) == KeyTable[i].hwcode ) {
		    idx = i;
		    process_hwcode(idx, hwcode);
		    break;
		}
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n");
	    fprintf(stderr,"Recommend removing other irda modules.\n");
	    exit(EXIT_FAILURE);
	}
    }  
}
